package uwe;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.Comment;
import org.eclipse.uml2.uml.Dependency;
import org.eclipse.uml2.uml.Feature;
import org.eclipse.uml2.uml.InstanceSpecification;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Stereotype;

/**
 * @author Marianne Busch <busch@pst.ifi.lmu.de>
 * @author Krzysztof Wolf <krzysztof.wolf.1980@googlemail.com>
 */
public class UWE2XACML {

	private static final String CALL_MSG = "\t-- CALL: ";
	public static final Logger logger = Helper.initLogger();

	private static final List<String> stereotypeList = new ArrayList<String>(6);
	private static final List<String> objectStereotypeList = new ArrayList<String>(2);

	// read comments only once
	static List<Comment> commentComments;

	static {
		stereotypeList.add("create");
		stereotypeList.add("read");
		stereotypeList.add("update");
		stereotypeList.add("delete");
		stereotypeList.add("readAll");
		stereotypeList.add("updateAll");

		objectStereotypeList.add("create");
		objectStereotypeList.add("delete");
	}

	/**
	 * Get UWE roles from UWE model.
	 * 
	 * @param basicRights
	 * @return List of roles
	 */
	public static Set<String> getUWERoles(Model basicRights) {
		logger.debug(CALL_MSG + "getUWERoles");
		List<InstanceSpecification> roles = Helper.findInstanceSpecificationsForObject(basicRights.getOwnedElements(), "Role");
		Set<String> roleStrings = new HashSet<String>();
		for (InstanceSpecification role : roles) {
			roleStrings.add(role.getLabel());
		}
		return roleStrings;
	}

	/**
	 * Get permissions for role from UWE model. Workaround: Links between roles
	 * have to be named "superRole::subRole". Workaround: States have to be
	 * named Role+"Area". (TODO: Remove workarounds)
	 * 
	 * @param basicRightsModel
	 * @param userModel
	 * @param contentModel
	 * @param navigationStates
	 * @param targetRole
	 * @return a set of Strings, action first (action__NUMBER, if more than one
	 *         similar with several constrains), separated by ": " (key) and
	 *         ", " (values). Values starting with "_constraint_" are
	 *         constraints.
	 */
	public static List<String> getPermissionsForRole(Model basicRightsModel, Model userModel, Model contentModel, Model navigationStates, String targetRole) {
		logger.info(CALL_MSG + "getPermissionsForRole(" + targetRole + ")");

		// group permissions according to the actions
		Map<String, List<String>> permissions = new HashMap<String, List<String>>();

		logger.info("targetrole:" + targetRole);

		// Navigation (TODO remove double target)
		final String NAVIGATE = "navigate";
		permissions.put(NAVIGATE, addElementToList(permissions.get(NAVIGATE), findAreaForRole(navigationStates, targetRole)));

		// BasicRights
		List<Dependency> dependencyList = Helper.filterDependenciesForClient(Helper.findDependencies(basicRightsModel.getOwnedElements(), stereotypeList), targetRole);

		// different slots in hashMap, if conditions are attached
		for (Dependency dependency : dependencyList) {
			List<Stereotype> appliedStereotypeList = dependency.getAppliedStereotypes();

			for (Stereotype stereotype : appliedStereotypeList) {
				final String stereotypeName = stereotype.getName();
				final String[] supplierPath = dependency.getSuppliers().get(0).getQualifiedName().split("::");
				final String directSupplier = supplierPath[(supplierPath.length - 1)];
				Set<Comment> constraints = getConditions(basicRightsModel, dependency);
				String actionKey = null;
				if (directSupplier.matches("\\p{Upper}.*")) { // dependency
																// to class
					if (stereotypeName.matches(".*All")) { // *.All
															// stereotypes

						// Search for the constrained class in the user or
						// the content model
						org.eclipse.uml2.uml.Class clazz = Helper.findClass(userModel.allOwnedElements(), directSupplier);
						if (clazz == null) {
							clazz = Helper.findClass(contentModel.allOwnedElements(), directSupplier);
						}
						Helper.printClass(clazz);
						List<Feature> propertyList = clazz.getFeatures();
						for (Feature feature : propertyList) {
							if (!isFeatureExcluded(dependency, feature)) {
								actionKey = addConstraints(permissions, stereotypeName.replaceAll("All", ""), constraints);
								permissions.put(actionKey, addElementToList(permissions.get(actionKey), directSupplier.toLowerCase() + "." + feature.getLabel()));
							}
						}

					} else { // single-action stereotypes
						if (objectStereotypeList.contains(stereotypeName)) {
							actionKey = addConstraints(permissions, stereotypeName, constraints);
							permissions.put(actionKey, addElementToList(permissions.get(actionKey), directSupplier.toLowerCase()));
						}
					}
				} else { // dependency to method / attribute
					actionKey = addConstraints(permissions, stereotypeName, constraints);
					final String supplierClass = supplierPath[supplierPath.length - 2];
					permissions.put(actionKey, addElementToList(permissions.get(actionKey), supplierClass.toLowerCase() + "." + directSupplier));
				}
			}
		}
		return permissionToString(permissions);
	}

	private static List<String> permissionToString(Map<String, List<String>> permissions) {
		// returns a set of Strings, action first, separated by ": " (key) and
		// ", " (values)
		List<String> result = new LinkedList<String>();
		for (String key : permissions.keySet()) {
			StringBuilder builder = new StringBuilder(key);
			builder.append(": ");
			for (String p : permissions.get(key)) {
				builder.append(p);
				builder.append(", ");
			}
			builder.delete(builder.length() - ", ".length(), builder.length());
			result.add(builder.toString());
		}
		return result;
	}

	private static String addConstraints(Map<String, List<String>> permissions, final String prefix, Set<Comment> constraints) {
		String actionKey = prefix;
		if (constraints != null && constraints.size() > 0) {
			Integer suffix = constraints.hashCode();
			actionKey = prefix + ((suffix != null) ? ("__" + suffix) : "");
			logger.fatal(actionKey + " " + permissions.get(actionKey));
			// don't add the same constraints again
			for (Comment c : constraints) {
				String constraint = "_constraint_" + c.getBody().substring("pre: ".length()).replace("<", "&lt;").replace(">", "&gt;");
				// MAYBE refactor this, because it's slow
				List<String> p = permissions.get(actionKey);
				if (p == null || !p.contains(constraint)) {
					permissions.put(actionKey, addElementToList(p, constraint));
				} else {
					// if one constraint was added, all others also were
					break;
				}
			}
		}
		return actionKey;
	}

	private static Set<Comment> getConditions(Model basicRightsModel, Dependency dependency) {
		List<String> stereotypeAuthList = new LinkedList<>();
		stereotypeAuthList.add("authorizationConstraint");
		if (commentComments == null) {
			commentComments = Helper.findComments(basicRightsModel.getOwnedElements(), stereotypeAuthList);
		}
		Set<Comment> result = new HashSet<>();
		for (Comment comment : commentComments) {
			if (comment.getAnnotatedElements().contains(dependency)) {
				result.add(comment);
			}
		}
		return result;
	}

	private static List<String> addElementToList(List<String> p, String addElement) {
		if (p == null) {
			p = new LinkedList<String>();
		}
		p.add(addElement);
		return p;
	}

	// MAYBE the following two functions should be refactored in order to remove
	// the necessity of naming links between roles
	public static Set<String> getSuperClassesForRole(Model basicRightsModel, String role) {
		logger.info(CALL_MSG + "getSuperClassesForRole(" + role + ")");
		Set<String> roleList = new HashSet<String>();
		List<InstanceSpecification> instanceSpecificationList = Helper.findInstanceSpecifications(basicRightsModel.getOwnedElements(), "::" + role);
		for (InstanceSpecification instanceSpecification : instanceSpecificationList) {
			String[] roles = instanceSpecification.getName().split("::");
			roleList.add(roles[0]);
			roleList.addAll(getSuperClassesForRole(basicRightsModel, roles[0]));
		}
		return roleList;
	}

	public static Set<String> getSubClassesForRole(Model basicRightsModel, String role) {
		logger.info(CALL_MSG + "getSubClassesForRole(" + role + ")");
		Set<String> roleList = new HashSet<String>();
		List<InstanceSpecification> instanceSpecificationList = Helper.findInstanceSpecifications(basicRightsModel.getOwnedElements(), role + "::");
		for (InstanceSpecification instanceSpecification : instanceSpecificationList) {
			String[] roles = instanceSpecification.getName().split("::");
			roleList.add(roles[1]);
			roleList.addAll(getSubClassesForRole(basicRightsModel, roles[1]));
		}
		return roleList;
	}

	@SuppressWarnings("unchecked")
	private static boolean isFeatureExcluded(Dependency dependency, Feature feature) {
		EList<Property> exceptList = (EList<Property>) Helper.findStereotypeProperty(dependency, "except");
		if (exceptList != null && !exceptList.isEmpty()) {
			for (Property exceptProperty : exceptList) {
				if (exceptProperty.getLabel().equals(feature.getLabel())) {
					return true;
				}
			}
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	private static String findAreaForRole(Model navigationStates, String roleName) {
		logger.info(CALL_MSG + "findAreaForRole(" + roleName + ")");
		StateMachine stateMachine = Helper.findStateHome(navigationStates.allOwnedElements()).containingStateMachine();
		List<State> stateList = Helper.findNavigationMenusStates(stateMachine.allOwnedElements(), Helper.findNavigationMenusClasses(navigationStates.allOwnedElements()));
		for (State state : stateList) {
			EList<InstanceSpecification> roleList = (EList<InstanceSpecification>) Helper.findStereotypeProperty(state, "roles");
			for (InstanceSpecification role : roleList) {
				if (role.getLabel().equals(roleName)) {
					logger.info("Found area: " + state.getLabel());
					return state.getLabel();
				}
			}
		}
		return null;
	}
}